#ifndef STATISTICS_H
#define STATISTICS_H

#include <string>
#include <unordered_map>
#include <climits>
#include <regex>

#include "SecurityRules.h"
#include "Utils.h"

typedef std::unordered_map<std::string, uint64_t> StringOccurrence;
typedef std::unordered_map<int, uint64_t> IntOccurrence;


/**
 * @brief minimal number of digits of a password, etc.
 */
struct minMax {
	/**
	 * @brief Update minima and maxima of all general data from analysed passwords
	 * with the analysed passwords
	 * @param m: other minMaxValue to merge with
	 */
	void updateMinMax(const minMax &m);
	bool operator==(const minMax& m) const;
	uint mindigit = UINT_MAX;
	uint maxdigit = 0;
	uint minlower = UINT_MAX;
	uint maxlower = 0;
	uint minupper = UINT_MAX;
	uint maxupper = 0;
	uint minspecial = UINT_MAX;
	uint maxspecial = 0;
};

struct Settings {
	std::string filename;			// input file
	std::string regex_str;			// select passwords you want
	bool withcount = false;			// if the file is formated [occurrences] [password]

	uint limitSimplemask = 12;		//minimal occurence fto be displayed
	uint limitAdvancedmask = 12;
	int hiderare = 0; 				// Hide low statistics
	int top = 10;					// Show only a top of statistics
	uint nbThread = 1;				// Number of usable threads, default 1
	std::string outfile_name;		// File where to write masks
	bool debug_enabled = false;		// Enable debug output
	SecurityRules sr = { 8, 0, 1, 1, 1 }; // Statistics on passwords matching these rules
};

struct Statistics {
	Statistics operator+(const Statistics& other) const;
	void operator+=(const Statistics& other);
	bool operator==(const Statistics& other) const;

	uint64_t total_counter = 0;
	uint64_t total_filter = 0;
	uint64_t nb_secure_passwords = 0;

	IntOccurrence length;
	StringOccurrence simplemasks;
	StringOccurrence advancedmasks;
	StringOccurrence charactersets;

	minMax minMaxValue;

	void show(std::ostream& os, const Settings& settings) const;
};

/**
 * @brief Print a line of statistics
 * @param occ: value of this statistic
 * @param mask: the statistic to display
 * @param total_counter:total number of password analysed
 * @param hiderare: flag to hide statistics under 1%
 */
template<typename Type>
std::ostream& readResult(std::ostream& os, const uint64_t& occ, const Type& mask, const uint64_t& total_counter, const int& hiderare) {
    if(occ == 0) return os;
    float perc = percentage(occ, total_counter);

    if (perc >= hiderare) {
        std::string value = std::to_string(perc);
        value = value.substr(0,5);

        os<< std::setw(40) << std::right << mask << ":  "
            << std::setw(5) << std::right << value << "%"
            << std::setw(5) << std::right << "(" << occ << ")";
    }
    return os;
}

/**
 * @brief Print an unordered map
 * @param stats: map to show
 * @param top: number of results to show
 * @param total_counter: number of finded passwords
 * @param hiderare: low statistics to hide
 * @param count: number of shown results
 */
template<typename Type>
std::ostream& showMap(std::ostream& os, const std::unordered_map<Type, uint64_t> & stats, const int & top, const uint64_t & total_counter, const int & hiderare) {
	int count = 0;
	std::multimap<uint64_t, Type, std::greater<uint64_t>> reverse = flip_map<Type>(stats);
	std::pair<uint64_t, Type> it;
	for(std::pair<uint64_t, Type> it : reverse) {
		readResult<Type>(os, it.first, it.second, total_counter, hiderare) << std::endl;
		count++;
		if (top != -1 && count == top) break;
	}

	if (count != top) {
		readResult<Type>(os, it.first, it.second, total_counter, hiderare) << std::endl;
	}
	return os;
}

std::ostream& operator<<(std::ostream& os, const Statistics& results);

#pragma omp declare reduction(+: Statistics : omp_out += omp_in ) initializer(omp_priv(omp_orig))

#endif // STATISTICS_H