/**
 * @file statsgen.cpp
 * @brief Creation of all methods from Statsgen
 *
 * @author Jean-Baptiste Jorand
 * @author Yannick Bass
 * Copyright (c) 2018 Jorand Jean-baptiste Bass Yannick
 * All rights reserved.
 *
 * Please see the attached LICENSE file for additional licensing information.
 */


#include <fstream>
#include <iomanip>
#include <algorithm>
#include <omp.h>


#include "Statsgen.h"
#include "Utils.h"

using namespace std;

int Statsgen::generate_stats() {
	int nbthreads;
	processed = 0;
	finished = false;
	results = Statistics();
	omp_set_num_threads(settings.nbThread);
	nblines = nbline_file(settings.filename);
	if (!nblines){
		cerr << "[ERROR] Empty file or not existing file" << endl;
		return 0;
	}

	started = true;
#pragma omp parallel reduction(+:results)
	{
#pragma omp single
		{
			nbthreads = omp_get_num_threads();
		}
		uint thread_id = omp_get_thread_num();
		ifstream inputfile(settings.filename);
		string line;
		int nbpasswords = 1;
		istringstream iss;
		for(uint numline = 0; numline < nblines; ++numline){
			getline(inputfile, line);
			if(settings.withcount){
				iss = istringstream(line);
				iss >> nbpasswords;
				iss >> line;
			}
			else {
				nbpasswords = 1;
			}
			if((numline % nbthreads) == thread_id){
				handle_password(line, nbpasswords, results);
#pragma omp atomic
				++processed;
			}
		}
	}

	finished = true;

	if (!results.total_counter) {
		cerr << "[ERROR] Empty file or not existing file" << endl;
		return 0;
	}

	return 1;
}

void Statsgen::print_stats() const {
	cout << settings.sr << endl;
	results.show(cout, settings);
}


pair<uint, uint> Statsgen::get_masks(const string& password, PasswordStats& c) const {
	char last_simplemask = 'a';
	uint sizeSimpleMask = 0;
	uint sizeAdvancedMask = 0;
	for(wchar_t letter: password){
		sizeAdvancedMask++;

		if (letter >= '0' && letter <= '9') {
			c.pol.digit++;
			c.advancedmask_string += "?d";
			if (last_simplemask != 'd') {
				sizeSimpleMask++;
				c.simplemask_string += "digit";
				last_simplemask = 'd';
			}
		}
		else if(letter >= 'a' && letter <= 'z') {
			c.pol.lower++;
			c.advancedmask_string += "?l";
			if (last_simplemask != 'l') {
				sizeSimpleMask++;
				c.simplemask_string += "lower";
				last_simplemask = 'l';
			}
		}
		else if(letter >= 'A' && letter <= 'Z') {
			c.pol.upper++;
			c.advancedmask_string += "?u";
			if (last_simplemask != 'u') {
				sizeSimpleMask++;
				c.simplemask_string += "upper";
				last_simplemask = 'u';
			}
		}
		else {
			c.pol.special++;
			c.advancedmask_string += "?s";

			if (last_simplemask != 's') {
				sizeSimpleMask++;
				c.simplemask_string += "special";
				last_simplemask = 's';
			}
		}
	}
	return make_pair(sizeSimpleMask, sizeAdvancedMask);
}

void Statsgen::handle_password(const string& password, const uint64_t& nbPasswords, Statistics& stats) const {
	stats.total_counter += nbPasswords;
	if(settings.regex_str.size() && !regex_match(password, std::regex(settings.regex_str))){
		return;
	}
	PasswordStats c;

	pair<uint, uint> masks = get_masks(password, c);

	if(c.pol.satisfies(settings.sr, password.size())){
		stats.nb_secure_passwords++;
	}

	if (masks.first > settings.limitSimplemask) {
		c.simplemask_string = "othermasks";
	}
	if (masks.second > settings.limitAdvancedmask) {
		c.advancedmask_string = "othermasks";
	}

	stats.total_filter += nbPasswords;
	stats.length[ password.size() ] += nbPasswords;
	stats.charactersets[ c.pol ] += nbPasswords;
	stats.simplemasks[ c.simplemask_string ] += nbPasswords;
	stats.advancedmasks[ c.advancedmask_string ] += nbPasswords;
	stats.minMaxValue.updateMinMax(c.pol);
}

bool Statsgen::operator==(const Statsgen& o) const {
	return results == o.results;
}

bool SecurityRules::operator==(const SecurityRules& o) const {
	return minLength == o.minLength
	&& minSpecial == o.minSpecial
	&& minDigit == o.minDigit
	&& minLower == o.minLower
	&& minUpper == o.minUpper
	;
}